--[[ 
	Thanks at skyboat for the original addon!
--]]

----------------------------
--      Declaration       --
----------------------------

Reslack = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceDB-2.0", "AceConsole-2.0")

----------------------------
--      Main Functions    --
----------------------------

local defaults = {
	AcceptResurrect = true,
	AcceptSummon = false,
	notice = false,
	AcceptReady = false,
	summondelay = false,
}

function Reslack:OnInitialize()
	self.options = {
		type="group",
		args = {
			resurrect = {
				name =	"Resurrect", type = "toggle",
				desc =	"Toggle the auto-resurrect function",
				get =	function()
							return self.db.profile.AcceptResurrect
						end,
						set = function(v)
							self.db.profile.AcceptResurrect = v
						end,
			},
			summon = {
				name =	"Summon", type = "toggle",
				desc =	"Toggle the auto-summon function",
				get =	function()
							return self.db.profile.AcceptSummon
						end,
						set = function(v)
							self.db.profile.AcceptSummon = v
						end,
			},
			ready = {
				name =	"Ready", type = "toggle",
				desc =	"Toggle the auto-ready function",
				get =	function()
							return self.db.profile.AcceptReady
						end,
						set = function(v)
							self.db.profile.AcceptReady = v
						end,
			},
			summondelay = {
				type = 'toggle',
				name = "SummonDelay",
				desc = "With this option enabled, automatic summons are delayed for 90 seconds.",
				get = function() return self.db.profile.summondelay end,
				set = function(v) self.db.profile.summondelay = v end,
	},
			

			notice = {
				name =	"Notice", type = "toggle",
				desc =	"Toggle notices to player",
				get =	function()
							return self.db.profile.notice
						end,
						set = function(v)
							self.db.profile.notice = v
						end,
			},
		},
	}
	
	-- Register everything
	self:RegisterDB("ReslackDB")
	self:RegisterDefaults('profile', defaults)
	self:RegisterChatCommand({ "/reslack", "/res" }, self.options )
end

function Reslack:OnEnable()
	self:RegisterEvent("READY_CHECK", "Ready")
	self:RegisterEvent("RESURRECT_REQUEST", "Resurrect")
	self:RegisterEvent("CONFIRM_SUMMON", "Summon")
	StaticPopupDialogs["CONFIRM_SUMMON"].OnCancel = function()
		if Reslack:IsEventScheduled("Reslack_Summon") then
			Reslack:CancelScheduledEvent("Reslack_Summon")
		end
	end

end

function Reslack:OnDisable()
	StaticPopupDialogs["CONFIRM_SUMMON"].OnCancel = function()
		return
	end
end


function Reslack:Resurrect()
	if not self.db.profile.AcceptResurrect then return end
	if (GetCorpseRecoveryDelay() == 0) then
		getglobal("StaticPopup1"):Hide()
		AcceptResurrect()
		if self.db.profile.notice then self:Print("Resurrection confirmed!") end
	else
		self:ScheduleEvent(function() getglobal("StaticPopup1"):Hide() AcceptResurrect() if self.db.profile.notice then self:Print("Ressurection confirmed!") end end, GetCorpseRecoveryDelay()+5)
	end
end

function Reslack:Summon()
if self.db.profile.AcceptSummon then
	if self.db.profile.summondelay and not self:IsEventScheduled("Reslack_Summon") then
		self:Print("Accepting summon in 90...")
		self:ScheduleEvent("Reslack_Summon", self.Confirm, 90, self)
	elseif not self:IsEventScheduled("Reslack_Summon") then
		self:Print("Accepting summon...")
		self:Confirm()
	end
end
end


function Reslack:Timeout()
	self:Print("Summon expired!")
	if self:IsEventScheduled("Reslack_Summon") then
		self:CancelScheduledEvent("Reslack_Summon")
	end
	if self:IsEventRegistered("PLAYER_REGEN_ENABLED") then
		self:UnregisterEvent("PLAYER_REGEN_ENABLED")
	end
end

function Reslack:LeaveCombat()
	self:Print("Left combat. Accepting summon in 10 seconds...")
	self:ScheduleEvent("module", self.Confirm, 10, self)
end

function Reslack:Confirm()
	if InCombatLockdown() then
		self:Print("In combat! Accepting summon after combat ends...")
		self:RegisterEvent("PLAYER_REGEN_ENABLED", self.AfterCombat)
		self:ScheduleEvent("Reslack_SummonTimeout", self.Timeout, GetSummonConfirmTimeLeft(), self)
		return
	end
	ConfirmSummon()
	StaticPopup_Hide("CONFIRM_SUMMON")
end

function Reslack:Ready()
	if not self.db.profile.AcceptReady then return end
	ReadyCheckFrame:Hide()
	self:ScheduleEvent(
		function()
			ConfirmReadyCheck(1)
			if self.db.profile.notice then self:Print("Ready check confirmed!") end
		end,
		1
	)
end
